# SPDX-License-Identifier: WTFPL
# SPDX-FileCopyrightText: 2024 Anna <cyber@sysrq.in>
# No warranty

from sortedcontainers import SortedSet
from repology_client.types import Package

from find_work.core.cli.options import MainOptions
from find_work.core.types.results import VersionBump
from find_work.plugins.repology.internal import collect_version_bumps
from find_work.plugins.repology.options import RepologyOptions


def test_collect_version_bumps():
    options = MainOptions()
    options.only_installed = False

    plugin_options = options.children["repology"] = RepologyOptions()
    plugin_options.repo = "example_linux"

    data = [
        {
            Package(
                repo="example_linux",
                visiblename="dev-util/examplepkg",
                version="1",
                status="outdated",
            ),
            Package(
                repo="example_bsd",
                visiblename="python-examplepkg",
                version="2",
                status="newest",
            ),
            Package(
                repo="example_macos",
                visiblename="py3-examplepkg",
                version="1",
                status="outdated",
            ),
        },
    ]

    expected = SortedSet([VersionBump("dev-util/examplepkg", "1", "2")])
    assert expected == collect_version_bumps(data, options)


def test_collect_version_bumps_multi_versions():
    options = MainOptions()
    options.only_installed = False

    plugin_options = options.children["repology"] = RepologyOptions()
    plugin_options.repo = "example_linux"

    data = [
        {
            Package(
                repo="example_linux",
                visiblename="dev-util/examplepkg",
                version="0",
                status="outdated",
            ),
            Package(
                repo="example_linux",
                visiblename="dev-util/examplepkg",
                version="1",
                status="outdated",
            ),
            Package(
                repo="example_bsd",
                visiblename="python-examplepkg",
                version="2",
                status="newest",
            ),
            Package(
                repo="example_macos",
                visiblename="py3-examplepkg",
                version="1",
                status="outdated",
            ),
        },
    ]

    expected = SortedSet([VersionBump("dev-util/examplepkg", "1", "2")])
    assert expected == collect_version_bumps(data, options)


def test_collect_version_bumps_multi_names():
    options = MainOptions()
    options.only_installed = False

    plugin_options = options.children["repology"] = RepologyOptions()
    plugin_options.repo = "example_linux"

    data = [
        {
            Package(
                repo="example_linux",
                visiblename="dev-util/examplepkg",
                version="0 pre-release",
                origversion="0",
                status="outdated",
            ),
            Package(
                repo="example_linux",
                visiblename="dev-util/examplepkg",
                version="1",
                status="outdated",
            ),
            Package(
                repo="example_linux",
                visiblename="dev-util/examplepkg-bin",
                version="1",
                status="outdated",
            ),
            Package(
                repo="example_bsd",
                visiblename="python-examplepkg",
                version="2",
                status="newest",
            ),
            Package(
                repo="example_macos",
                visiblename="py3-examplepkg",
                version="1",
                status="outdated",
            ),
        },
    ]

    expected = SortedSet([
        VersionBump("dev-util/examplepkg", "1", "2"),
        VersionBump("dev-util/examplepkg-bin", "1", "2"),
    ])
    assert expected == collect_version_bumps(data, options)



# SPDX-License-Identifier: WTFPL
# SPDX-FileCopyrightText: 2025 Anna <cyber@sysrq.in>
# No warranty

"""
Results reporter for Repology results, such as repository problems.
"""

from typing import Any

from pydantic import validate_call
from repology_client.types import Problem

from find_work.core.reporters import AbstractReporter


class ConsoleProblemReporter(AbstractReporter[Problem]):
    reporter_name = "console"
    result_type = Problem

    def add_result(self, item: Problem) -> None:
        match item.type:
            case "homepage_dead":
                url = item.data.get("url", "URL is missing")
                status = item.data.get("status", 0)
                self.options.echo("\t", nl=False)
                self.options.echo(f"Homepage link <{url}> is dead ()")



# SPDX-License-Identifier: WTFPL
# SPDX-FileCopyrightText: 2024 Anna <cyber@sysrq.in>
# No warranty

"""
Repology subcommand options.
"""

from collections.abc import Sequence

from find_work.core.cli.options import OptionsBase
from find_work.core.types import VersionPart


class OutdatedCmdOptions(OptionsBase):
    """
    Options for ``outdated`` command.
    """

    #: Version part filter.
    version_part: VersionPart | None = None

    @property
    def attr_order(self) -> Sequence[str]:
        return tuple()


class RepologyOptions(OptionsBase):
    """
    Options for Repology subcommands.
    """

    #: Repository name.
    repo: str = ""

    @property
    def attr_order(self) -> Sequence[str]:
        return ["repo"]



# SPDX-License-Identifier: WTFPL
# SPDX-FileCopyrightText: 2024 Anna <cyber@sysrq.in>
# No warranty

"""
Internal functions that don't depend on any CLI functionality.
"""

from collections.abc import Collection, Mapping, Set

import gentoopm
import repology_client
import repology_client.exceptions
from gentoopm.basepm.atom import PMAtom
from pydantic import validate_call
from repology_client.types import Package
from sortedcontainers import SortedSet

from find_work.core.cli.options import MainOptions
from find_work.core.types.results import VersionBump
from find_work.core.utils import aiohttp_session

from find_work.plugins.repology.options import RepologyOptions

PackageSet = Set[Package]
ProjectsMapping = Mapping[str, PackageSet]


@validate_call
async def fetch_outdated(options: MainOptions) -> ProjectsMapping:
    plugin_options = RepologyOptions.model_validate(
        options.children["repology"]
    )

    filters: dict = {}
    if options.maintainer:
        filters["maintainer"] = options.maintainer
    if options.category:
        filters["category"] = options.category

    async with aiohttp_session() as session:
        return await repology_client.get_projects(inrepo=plugin_options.repo,
                                                  outdated="on", count=5_000,
                                                  session=session, **filters)


def collect_version_bumps(data: Collection[PackageSet],
                          options: MainOptions) -> SortedSet[VersionBump]:
    plugin_options = RepologyOptions.model_validate(
        options.children["repology"]
    )
    pm = gentoopm.get_package_manager()

    result: SortedSet[VersionBump] = SortedSet()
    for packages in data:
        latest_pkgs: dict[str, PMAtom] = {}  # latest in repo, not across repos!
        new_version: str | None = None

        for pkg in packages:
            if pkg.status == "outdated" and pkg.repo == plugin_options.repo:
                # "pkg.version" can contain spaces, better avoid it!
                origversion = pkg.origversion or pkg.version
                atom = pm.Atom(f"={pkg.visiblename}-{origversion}")

                latest = latest_pkgs.get(pkg.visiblename)
                if latest is None or atom.version > latest.version:
                    latest_pkgs[pkg.visiblename] = atom
            elif pkg.status == "newest":
                new_version = pkg.version

        for latest in latest_pkgs.values():
            if not (options.only_installed and latest.key not in pm.installed):
                result.add(VersionBump(str(latest.key), str(latest.version),
                                       new_version or "(unknown)"))
    return result



# SPDX-License-Identifier: WTFPL
# SPDX-FileCopyrightText: 2024 Anna <cyber@sysrq.in>
# No warranty

"""
Personal advice utility for Gentoo package maintainers: Repology plugin
"""

import click
from click_aliases import ClickAliasedGroup

from find_work.core.cli.options import MainOptions
from find_work.core.cli.plugins import cli_hook_impl

import find_work.plugins.repology.cli as plugin_cli
from find_work.plugins.repology.options import RepologyOptions


@cli_hook_impl
def attach_base_command(group: ClickAliasedGroup) -> None:
    group.add_command(plugin_cli.repology, aliases=["rep", "r"])


@cli_hook_impl
def setup_base_command(options: MainOptions) -> None:
    if "repology" not in options.children:
        options.children["repology"] = RepologyOptions()


@cli_hook_impl
def get_command_by_name(command: str) -> click.Command | None:
    plug_name, cmd_name = command.split(":")[:2]
    if plug_name == "repology":
        match cmd_name:
            case "outdated":
                return plugin_cli.outdated
    return None



# SPDX-License-Identifier: WTFPL
# SPDX-FileCopyrightText: 2024 Anna <cyber@sysrq.in>
# No warranty

"""
Implementation of CLI commands for the Repology plugin.
"""

import asyncio

import click
from click_aliases import ClickAliasedGroup
from pydantic import TypeAdapter, validate_call

from find_work.core.cache import (
    read_raw_json_cache,
    write_raw_json_cache,
)
from find_work.core.cli.messages import Status, Result
from find_work.core.cli.options import MainOptions
from find_work.core.cli.widgets import ProgressDots
from find_work.core.types import VersionPart
from find_work.core.types.results import VersionBump

from find_work.plugins.repology.options import (
    OutdatedCmdOptions,
    RepologyOptions,
)


@validate_call
async def _outdated(options: MainOptions) -> None:
    import repology_client.exceptions
    from find_work.plugins.repology.internal import (
        ProjectsMapping,
        collect_version_bumps,
        fetch_outdated,
    )

    plugin_options = RepologyOptions.model_validate(
        options.children["repology"]
    )
    cmd_options = OutdatedCmdOptions.model_validate(
        plugin_options.children["outdated"]
    )
    dots = ProgressDots(options.verbose)

    data: ProjectsMapping
    with dots(Status.CACHE_READ):
        raw_data = read_raw_json_cache(options.breadcrumbs)
    if raw_data:
        with dots(Status.CACHE_LOAD):
            data = TypeAdapter(ProjectsMapping).validate_json(raw_data)
    else:
        try:
            with dots("Fetching data from Repology API"):
                data = await fetch_outdated(options)
        except repology_client.exceptions.EmptyResponse:
            return options.exit(Result.EMPTY_RESPONSE)
        with dots(Status.CACHE_WRITE):
            adapter: TypeAdapter[ProjectsMapping] = TypeAdapter(ProjectsMapping)
            raw_json = adapter.dump_json(data, exclude_none=True)
            write_raw_json_cache(raw_json, options.breadcrumbs)

    no_work = True
    with options.get_reporter_for(VersionBump) as reporter:
        for bump in collect_version_bumps(data.values(), options):
            if (
                cmd_options.version_part is None
                or bump.changed(cmd_options.version_part)
            ):
                reporter.add_result(bump)
                no_work = False

    if no_work:
        return options.exit(Result.NO_WORK)


@click.group(cls=ClickAliasedGroup,
             epilog="See `man find-work-repology` for the full help.")
@click.option("-r", "--repo", metavar="NAME", required=True,
              help="Repository name on Repology.")
@click.pass_obj
def repology(options: MainOptions, repo: str = "", *,
             indirect_call: bool = False) -> None:
    """
    Use Repology to find work.
    """

    options.breadcrumbs.feed("repology")

    plugin_options = RepologyOptions.model_validate(
        options.children["repology"]
    )

    if not indirect_call:
        plugin_options.repo = repo

    for key in plugin_options.attr_order:
        options.breadcrumbs.feed_option(key, plugin_options[key])


@repology.command(aliases=["out", "o"])
@click.option("-F", "--filter", "version_part",
              type=click.Choice(["major", "minor", "patch"]),
              help="Version part filter.")
@click.pass_context
def outdated(ctx: click.Context, version_part: VersionPart | None = None, *,
             init_parent: bool = False) -> None:
    """
    Find outdated packages.
    """

    options: MainOptions = ctx.obj
    if init_parent:
        ctx.invoke(repology, indirect_call=True)

    options.breadcrumbs.feed("outdated")

    plugin_options = RepologyOptions.model_validate(
        options.children["repology"]
    )
    cmd_options = plugin_options.children["outdated"] = OutdatedCmdOptions()
    cmd_options.version_part = version_part

    asyncio.run(_outdated(options))
